#include <u.h>
#include <libc.h>
#include "xml.h"

static void
freestr(char *s, uchar fill)
{
	if(xml_debug)
		memset(s, fill, strlen(s));
	free(s);
}

static void
freemem(void *v, uchar fill, int len)
{
	if(xml_debug)
		memset(v, fill, len);
	free(v);
}


static void
xfree(Elem *ep)
{
	Elem *te;
	Attr *ap, *ta;

	while(ep){
		if(ep->name)
			freestr(ep->name, 0xff);
		if(ep->pcdata)
			freestr(ep->pcdata, 0xfe);
		ap = ep->attrs;
		while(ap){
			if(ap->name)
				freestr(ap->name, 0xfd);
			if(ap->value)
				freestr(ap->value, 0xfc);
			ta = ap->next;
			if(ap)
				freemem(ap, 0xfb, sizeof(Attr));
			ap = ta;
		}
		if(ep->child)
			xfree(ep->child);

		te = ep->next;
		freemem(ep, 0xfa, sizeof(Elem));
		ep = te;
	}
}

static void
freetree(Xml *xp)
{
	xfree(xp->root);
	free(xp->doctype);
}

/********************/

#define Roundup(x, g)	(((x) + (unsigned)(g-1)) & ~((unsigned)(g-1)))

struct xml_tree {
	xml_tree *left;
	xml_tree *right;
	char *str;
	int hits;
};

struct xml_block {
	xml_block *next;
	char *free;
	char *end;
};

static int Strdups, Commons, Unique, Memblocks;

static void *
getmem(Xml *xp, int len)
{
	int sz;
	xml_block *b;
	char *ret;

	len = Roundup(len, sizeof(long long));

	sz = xp->alloc.blksiz;		/* shorthand */
	b = xp->alloc.active;

	if(len > sz)
		sysfatal("xml heap: object larget than blocksize (%d > %d)\n", len, sz);

	if(xp->alloc.active == nil || b->free + len >= b->end){
		Memblocks++;
		b = mallocz(sizeof(xml_block) + sz, 0);
		b->free = (char *)&b[1];
		b->end = (char *)&b->free[sz];

		b->next = xp->alloc.active;
		xp->alloc.active = b;
	}

	ret = b->free;
	b->free += len;
	return ret;
}

static xml_tree *
lookadd(Xml *xp, xml_tree *t, char *str, xml_tree **match)
{
	int n;

	if(t == nil){
		Unique++;
		t = getmem(xp, sizeof(xml_tree) + strlen(str)+1);
		t->left = nil;
		t->right = nil;
		t->str = (char *)&t[1];
		strcpy(t->str, str);
		*match = t;
		t->hits = 1;
		return t;
	}

	if((n = strcmp(str, t->str)) == 0){
		*match = t;
		t->hits++;
	}
	if(n < 0)
		t->left = lookadd(xp, t->left, str, match);
	if(n > 0)
		t->right = lookadd(xp, t->right, str, match);
	return t;
}

static void
heapfree(Xml *xp)
{
	xml_block *b, *n;

	for(b = xp->alloc.active; b; b = n){
		n = b->next;
		if(xml_debug)
			memset(b, 0x7e, xp->alloc.blksiz);
		free(b);
	}

}

static void
dumpstats(Xml *)
{
	fprint(2, "total=%d common=%d -> unique=%d rare=%d memblocks=%d\n",
		Strdups, Commons, Unique, Strdups - Commons, Memblocks);
}

char *
xml_strdup(Xml *xp, char *str, int iscommon)
{
	char *s;
	xml_tree *t;

	if(xp->alloc.blksiz <= 0)
		return strdup(str);

	Strdups++;
	if(iscommon){
		Commons++;
		xp->alloc.root = lookadd(xp, xp->alloc.root, str, &t);
		return t->str;
	}

	s = getmem(xp, strlen(str)+1);
	return strcpy(s, str);
}

void *
xml_calloc(Xml *xp, long n, long m)
{
	void *v;

	if(xp->alloc.blksiz <= 0)
		return calloc(n, m);

	v = getmem(xp, n * m);
	memset(v, 0, n * m);
	return v;
}

void
xml_free(Xml *xp)
{
	if(xp->alloc.blksiz <= 0)
		freetree(xp);
	else
		heapfree(xp);
	free(xp);
}

